"""Toolbox for phase response curves measured by finite perturbations
"""

import numpy as np
from PyDSTool import Pointset, Model, embed
from PyDSTool.Trajectory import pointset_to_traj

#from pylab import figure, plot

#__all__ = []

# -------------------------------------------------

def one_period_traj(model, ev_name, ev_t_tol, ev_norm_tol, T_est,
                    verbose=False, initial_settle=6, restore_old_ics=False):
    """
    Utility to extract a single period of a limit cycle of the model using forward
    integration, up to a tolerance given in terms of both the period and the norm of the
    vector of variables in the limit cycle at the period endpoints.

    Requires a non-terminal event in the model that is detected exactly once per period.
    Assumes model initial conditions are already in domain of attraction for limit cycle.

    T_est is an initial estimate of period.
    The model argument can be an instance of a Generator class or Model class.

    Returned trajectory will have name 'one_period'.
    """
    if not isinstance(model, Model.Model):
        # temporarily embed into a model object
        model = embed(model)
    trajname = '_test_period_'
    old_ics = model.query('ics')
    settle = initial_settle
    tries = 1
    success = False
    while not success and tries < 8:
        model.compute(trajname=trajname, tdata=[0,T_est*(settle+0.2)], force=True)
        evts = model.getTrajEventTimes(trajname, ev_name)
        all_evs = model.getTrajEventTimes(trajname)
        if len(evts) <= 2:
            raise RuntimeError("Not enough events found")
        ref_ic = model(trajname, evts[-1])
        t_check = 10000*np.ones((tries,),float)
        norm_check = 10000*np.ones((tries,),float)
        T = np.zeros((tries,),float)
        look_range = range(1, min((tries+1, len(evts))))
        if verbose:
            print "\n Tries: ", tries, "\n"
        for lookback in look_range:
            try:
                d_evts = [evts[i]-evts[i-lookback] for i in \
                                    range(lookback, len(evts))]
            except KeyError:
                # no more events left to look back at
                break
            else:
                prev_val = model(trajname, evts[-(1+lookback)])
                t_check[lookback-1] = abs(d_evts[-1]-d_evts[-2])
                norm_check[lookback-1] = np.linalg.norm(ref_ic-prev_val)
                T[lookback-1] = d_evts[-1]
        T_est = T[0]
        t_ix = np.argmin(t_check)
        n_ix = np.argmin(norm_check)
        ix1 = min((t_ix, n_ix))
        ix2 = max((t_ix, n_ix))
        if verbose:
            print t_check, norm_check, T
            print ix1, ix2
        if t_check[ix1] < ev_t_tol and norm_check[ix1] < ev_norm_tol:
            success = True
            T_final = T[ix1]
        elif ix1 != ix2 and t_check[ix2] < ev_t_tol and norm_check[ix2] < ev_norm_tol:
            success = True
            T_final = T[ix2]
        else:
            tries += 1
            settle = tries*2
            model.set(ics = ref_ic)
    if success:
        model.set(ics=ref_ic, tdata=[0, T_final])
        model.compute(trajname='one_period', force=True)
        ref_traj = model['one_period']
        # insert the ON event at beginning of traj
        ref_traj.events[ev_name] = Pointset(indepvararray=[0],
                                        coordarray=np.array([ref_ic.coordarray]).T,
                                        coordnames=ref_ic.coordnames)
        ref_pts = ref_traj.sample()
        # restore old ICs
        if restore_old_ics:
            model.set(ics=old_ics)
        return ref_traj, ref_pts, T_final
    else:
        raise RuntimeError("Failure to converge after 80 iterations")




def finitePRC(model, ref_traj_period, evname, pertcoord, pertsize=0.05, settle=5, verbose=False):
    ref_pts = ref_traj_period.sample()
    ref_ts = ref_pts.indepvararray
    T = ref_ts[-1]-ref_ts[0]
    PRCvals = []
    for i, t0 in enumerate(ref_ts):
        if verbose:
            print i, "of", len(ref_ts), ": t0 = ", t0, "of", T
        ic = ref_pts[i]
        ic[pertcoord] += pertsize
        model.set(ics=ic, tdata=[0,settle*T+t0])
        model.compute(trajname='pert', force=True)
        evts = model.getTrajEventTimes('pert', evname)
        val = np.mod(evts[-1]+t0, T)/T
        if val > 0.5:
            val = 1-val
        PRCvals.append(val)
    return Pointset(coordarray=[PRCvals], coordnames=['D_phase'], indepvararray=ref_ts, indepvarname='t')
