import numpy as np


class ScaledLineSearch(object):
    """
    A simple line search, takes a point, adds a step and returns it
    Scales step according to given scales of the parameters and ignores
    *magnitude* of gradient.

    (in early development and experimental only at this point)
    """
    def __init__(self, max_step=1, step_fac=3, **kwargs):
        """
        Needs to have :
        - nothing
        Can have :
        - max_step: a step modifier, a scalar or vector to restrict step size
        in each direction (default 1.)
        """
        self.maxStepSize = max_step
        self.stepFac = step_fac
        if np.isscalar(max_step):
            self.basis = None
            self.dim = None
        else:
            self.dim = len(max_step)
            self.basis = np.identity(self.dim)
        try:
            self.use_dirs = kwargs['use_directions']
        except KeyError:
            self.use_dirs = np.zeros((self.dim,),float)


    def __call__(self, origin, state, **kwargs):
        """
        Returns a good candidate
        Parameters :
        - origin is the origin of the search
        - state is the state of the optimizer
        """
        fun = kwargs['function']
        direction = state['direction']/np.linalg.norm(state['direction'])
        for pos, d in enumerate(direction):
            use_dir = self.use_dirs[pos]
            if use_dir * d < 0:
                # directions don't match so don't move in this direction
                direction[pos] = 0
        state['direction'] = direction
        maxStepSize = self.maxStepSize
        if np.isscalar(maxStepSize):
            stepSize = maxStepSize
        else:
            stepfacs = np.zeros(self.dim)
            for d in range(self.dim):
                # explicit loop so as to catch any ZeroDivisionErrors
                try:
                    stepfacs[d] = abs(maxStepSize[d] / direction[d])
                except ZeroDivisionError:
                    # Direction is orthogonal to this parameter direction,
                    # so ensure won't choose this as the minimum step size
                    step[d] = Inf
            # Stop stepping with giant sizes if direction vector has strong
            # separation of scales
            stepSize = min(stepfacs)
#            print "direction = ", direction
#            print "step = ", step
        init_step = stepSize
        i = 1
        old_value = state['old_value']
        not_done = True
        while not_done:
            print "\nLinestep: i =", i, "step size =", stepSize, "direction =", direction
            p = origin + i * stepSize * direction
            print "p = ", p
            new_value = fun(p)
            if new_value < old_value:
                i += 1
                old_value = new_value
            else:
                if i == 1:
                    # don't shrink step size to be less than 1/200 of initial
                    if stepSize*200 < init_step:
                        not_done = False
                        p = origin
                    else:
                        stepSize /= self.stepFac
                else:
                    not_done = False
                    p = origin + (i-1) * stepSize * direction
        state['alpha_step'] = stepSize
        return p
